import inspect
from typing import Any

from ._exceptions import (
    _UnsetExitError,
    SYMBOL,
)
from ._switch_var import _is_mult_vars


def _store_org_value(self, label: str, index: int, org_value: Any) -> None:
    target_index = self._org_value[label][index] 

    if target_index is None:
        self._org_value[label][index] = org_value
        return
    elif not isinstance(target_index, list):
        self._org_value[label][index] = [target_index]
    elif not _is_mult_vars(target_index):
        self._org_value[label][index] = [target_index]

    if isinstance(org_value, (list, tuple)):
      for v in org_value:
          self._org_value[label][index].append(v)
    else:
      self._org_value[label][index].append(org_value)

def _update_var_value(
    self, 
    var_ref: tuple[Any] | list[tuple[Any]], update_value: Any,
) -> None:
    # var_ref can be:
    # - (lineno, var_name) for globals
    # - (lineno, attr_name, instance) for class attributes
    # - a list of either form

    if isinstance(var_ref, list):
        # When multiple variables are provided, a list is used
        for value in var_ref:
            if len(value) == 3:
                cur_value = var_ref[2].__dict__[var_ref[1]]

                if cur_value == update_value:
                    continue

                setattr(value[2], value[1], update_value)
            else:
                cur_value = self._frame.f_globals[value[1]]

                if cur_value == update_value:
                    continue 

                self._frame.f_globals[value[1]] = update_value   
    else:     
        if len(var_ref) == 3:
            cur_value = var_ref[2].__dict__[var_ref[1]]

            if cur_value == update_value:
                return
            
            setattr(var_ref[2], var_ref[1], update_value)
        else:  
            cur_value = self._frame.f_globals[var_ref[1]]
            
            if cur_value == update_value:
                return
            
            self._frame.f_globals[var_ref[1]] = update_value

def _check_label_flag(self, label: str, cond: str | None) -> None:
    target_func = "set_trigger"

    name = label.lstrip(SYMBOL)
    self._check_exist_label(name)

    if self._disable_label[name] or self._trigger_flag[name]:
        return
    
    if cond is not None:
        self._get_target_frame(target_func)

        if not self._ensure_safe_cond(cond):
            return
        
    self._trigger_flag[name] = True
    
    self._label_has_var(name, target_func, False)

    if self.debug:
        self._get_target_frame(target_func)
        self._print_flag_debug(name, "active", clear_fram=False)

def _label_has_var(
    self, label: str, called_func: str, to_org: bool,
) -> None:
    if self._var_list[label] is None:
        return
    
    if to_org:
        update_value = self._org_value[label]
    else:
        update_value = self._new_value[label]

    self._get_target_frame(called_func)

    # Restore all variables to their original or new values 
    for i in range(len(self._var_list[label])):
        var_ref = self._var_list[label][i]

        if var_ref is None:
            continue
        elif isinstance(var_ref, list):
            for i_2, val in enumerate(var_ref):
                if not isinstance(update_value[i], (list, tuple)):
                    self._update_var_value(val, update_value[i])
                    continue            

                self._update_var_value(val, update_value[i][i_2])
        else:
            self._update_var_value(var_ref, update_value[i])      

def _get_target_frame(
        self, target_name: str | list[str, str], has_exit: bool=False,
) -> None:
   if self._frame is not None:
      return

   frame = inspect.currentframe()

   while frame:
      if has_exit:
          if frame.f_code.co_name == "<module>":
              raise _UnsetExitError()
          elif frame.f_code.co_name == target_name:
              return
      elif isinstance(target_name, list):
          if frame.f_code.co_name in target_name:
              self._frame = frame.f_back
              return
      elif frame.f_code.co_name == target_name:
         self._frame = frame.f_back
         return     
      
      frame = frame.f_back

def _clear_frame(self) -> None:
   # to prevent memory leak by releasing the frame reference
   self._frame = None
   self._lineno = None




import ast
from typing import Any

from ._err_handler import _count_symbol
from ._exceptions import SYMBOL

def _init_or_not(
        self, 
        label: str, index: int, data: dict[str, Any], 
        org_val: Any, arg_type: ast.AST,
) -> bool:
    if len(data) == 1:
        shared_index = index
    else:
        shared_index = None   

    mult_flag = _is_mult_vars(org_val)
    
    if mult_flag:
        target_val = org_val[0]
    else:
        target_val = org_val

    if (
        self._var_list[label][index] is None or 
        self._is_new_var(label, index, target_val)
    ):  
        self._get_target_frame(["switch_var", "alter_var"]) # Will change it after beta
        self._lineno = self._frame.f_lineno     

        # Initial process to store argument variables 
        self._init_arg_list(data, arg_type, shared_index)
 
        if shared_index is None:
            for key, val in data.items():
                name = key.lstrip(SYMBOL)
                index = _count_symbol(key)

                if mult_flag:
                    for v in val:
                        if self._org_value[name][index] is None:
                            self._store_org_value(name, index, v)
                else:
                    self._store_org_value(name, index, val)
                                
                self._find_match_var(name, index)
        else:
            self._store_org_value(label, index, data[label])
            self._find_match_var(label, index)

        return True
    
    return False

def _is_new_var(self, label: str, index: int, value: Any) -> bool:
    # The new value to be set
    if self._trigger_flag[label]:
        if self._new_value[label][index] == value:
            return False
        
        return True  
        
    # The original value to be set
    if isinstance(value, list):
        if self._org_value[label][index] == value:
            return False
    elif isinstance(value, tuple):
        if tuple(self._org_value[label][index]) == value:
            return False
    elif self._org_value[label][index] == value:
            return False
    
    return True

# Will update list type annotation later

def _find_match_var(
        self, label: str=None, index: int=None, 
        target_ref: tuple[str, ...]=None, init: bool=False,
) -> None | bool:
    # Used to retrieve the original value for `revert()`
    # or check if `target_ref` is already registered for a specific label.

    if not init:
      # it can be a list or tuple
      var_ref = self._var_list[label][index]
    else:
        # it always be a tuple
        var_ref = target_ref

    stop_flag = False

    # `value` is a list
    for key, value in self._var_list.items(): 
        if stop_flag and not init:
            return
        elif stop_flag and init:
            return False
        elif init and key != label:
            continue

        if key == label:
            stop_flag = True

        # `list_val` is a list or tuple or None
        for i, list_val in enumerate(value):
            if list_val is None:
                continue
            elif isinstance(list_val, tuple) and isinstance(var_ref, tuple):
                if self._is_ref_match(list_val, var_ref):
                    if not init:
                        self._org_value[label][index] = self._org_value[key][i]     
                    else:
                        return True

            elif isinstance(list_val, tuple) and isinstance(var_ref, list):
                for i_2, ref_v in enumerate(var_ref):
                    if self._is_ref_match(list_val, ref_v):
                        self._org_value[label][index][i_2] = (
                            self._org_value[key][i]
                        )
            elif isinstance(list_val, list) and isinstance(var_ref, tuple):
                for i_2, l_v in enumerate(list_val):
                    if self._is_ref_match(l_v, var_ref):
                        if not init:
                            self._org_value[label][index] = (
                                self._org_value[key][i][i_2]
                            )
                        else:
                            return True
            else:
                for ref_i, ref_val in enumerate(var_ref):
                    for list_i, v in enumerate(list_val):
                        if self._is_ref_match(v, ref_val):                           
                            self._org_value[label][index][ref_i] = (
                                self._org_value[key][i][list_i]
                            )

def _is_ref_match(
        self,
        list_val: tuple[str, ...], target_val: tuple[str, ...],
) -> bool:
    if len(list_val) == 2 and len(target_val) == 2:
        return list_val[1] == target_val[1]

    elif len(list_val) == 3 and len(target_val) == 3:
        return list_val[1:] == target_val[1:]
    

def _is_mult_vars(value: Any) -> bool:
  if isinstance(value, (list, tuple)):
    if len(value) > 1:
      return True
    
    return False
  
  return False